# flake8: noqa:

import bmy

import os
from myrrh.utils.myrrh_test_init import setUp

setUp(f"bmy.{os.path.basename(__file__)}")

with bmy.select():
    from mlib.py import io

    open = io.open

with bmy.select():
    from mlib.py import os

import copy
import pickle

with bmy.select():
    from mlib.py import platform
    from mlib.py import subprocess
    from mlib.py import sys

import unittest
from unittest import mock

with bmy.select():
    from mlib.py.test import support
    from mlib.py.test.support import os_helper

FEDORA_OS_RELEASE = """\
NAME=Fedora
VERSION="32 (Thirty Two)"
ID=fedora
VERSION_ID=32
VERSION_CODENAME=""
PLATFORM_ID="platform:f32"
PRETTY_NAME="Fedora 32 (Thirty Two)"
ANSI_COLOR="0;34"
LOGO=fedora-logo-icon
CPE_NAME="cpe:/o:fedoraproject:fedora:32"
HOME_URL="https://fedoraproject.org/"
DOCUMENTATION_URL="https://docs.fedoraproject.org/en-US/fedora/f32/system-administrators-guide/"
SUPPORT_URL="https://fedoraproject.org/wiki/Communicating_and_getting_help"
BUG_REPORT_URL="https://bugzilla.redhat.com/"
REDHAT_BUGZILLA_PRODUCT="Fedora"
REDHAT_BUGZILLA_PRODUCT_VERSION=32
REDHAT_SUPPORT_PRODUCT="Fedora"
REDHAT_SUPPORT_PRODUCT_VERSION=32
PRIVACY_POLICY_URL="https://fedoraproject.org/wiki/Legal:PrivacyPolicy"
"""

UBUNTU_OS_RELEASE = """\
NAME="Ubuntu"
VERSION="20.04.1 LTS (Focal Fossa)"
ID=ubuntu
ID_LIKE=debian
PRETTY_NAME="Ubuntu 20.04.1 LTS"
VERSION_ID="20.04"
HOME_URL="https://www.ubuntu.com/"
SUPPORT_URL="https://help.ubuntu.com/"
BUG_REPORT_URL="https://bugs.launchpad.net/ubuntu/"
PRIVACY_POLICY_URL="https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
VERSION_CODENAME=focal
UBUNTU_CODENAME=focal
"""

TEST_OS_RELEASE = r"""
# test data
ID_LIKE="egg spam viking"
EMPTY=
# comments and empty lines are ignored

SINGLE_QUOTE='single'
EMPTY_SINGLE=''
DOUBLE_QUOTE="double"
EMPTY_DOUBLE=""
QUOTES="double\'s"
SPECIALS="\$\`\\\'\""
# invalid lines
=invalid
=
INVALID
IN-VALID=value
IN VALID=value
"""


class PlatformTest(unittest.TestCase):
    def clear_caches(self):
        platform._platform_cache.clear()
        platform._sys_version_cache.clear()
        platform._uname_cache = None
        platform._os_release_cache = None

    def test_architecture(self):
        res = platform.architecture()

    @os_helper.skip_unless_symlink
    @support.requires_subprocess()
    def test_architecture_via_symlink(self):  # issue3762
        with support.PythonSymlink() as py:
            cmd = "-c", "import platform; print(platform.architecture())"
            self.assertEqual(py.call_real(*cmd), py.call_link(*cmd))

    def test_platform(self):
        for aliased in (False, True):
            for terse in (False, True):
                res = platform.platform(aliased, terse)

    def test_system(self):
        res = platform.system()

    def test_node(self):
        res = platform.node()

    def test_release(self):
        res = platform.release()

    def test_version(self):
        res = platform.version()

    def test_machine(self):
        res = platform.machine()

    def test_processor(self):
        res = platform.processor()

    def setUp(self):
        self.save_version = sys.version
        self.save_git = sys._git
        self.save_platform = sys.platform

    def tearDown(self):
        sys.version = self.save_version
        sys._git = self.save_git
        sys.platform = self.save_platform

    def test_sys_version(self):
        # Old test.
        for input, output in (
            (
                "2.4.3 (#1, Jun 21 2006, 13:54:21) \n[GCC 3.3.4 (pre 3.3.5 20040809)]",
                (
                    "CPython",
                    "2.4.3",
                    "",
                    "",
                    "1",
                    "Jun 21 2006 13:54:21",
                    "GCC 3.3.4 (pre 3.3.5 20040809)",
                ),
            ),
            (
                "IronPython 1.0.60816 on .NET 2.0.50727.42",
                ("IronPython", "1.0.60816", "", "", "", "", ".NET 2.0.50727.42"),
            ),
            (
                "IronPython 1.0 (1.0.61005.1977) on .NET 2.0.50727.42",
                ("IronPython", "1.0.0", "", "", "", "", ".NET 2.0.50727.42"),
            ),
            (
                "2.4.3 (truncation, date, t) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "date t", "GCC"),
            ),
            (
                "2.4.3 (truncation, date, ) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "date", "GCC"),
            ),
            (
                "2.4.3 (truncation, date,) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "date", "GCC"),
            ),
            (
                "2.4.3 (truncation, date) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "date", "GCC"),
            ),
            (
                "2.4.3 (truncation, d) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "d", "GCC"),
            ),
            (
                "2.4.3 (truncation, ) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "", "GCC"),
            ),
            (
                "2.4.3 (truncation,) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "", "GCC"),
            ),
            (
                "2.4.3 (truncation) \n[GCC]",
                ("CPython", "2.4.3", "", "", "truncation", "", "GCC"),
            ),
        ):
            # branch and revision are not "parsed", but fetched
            # from sys._git.  Ignore them
            (
                name,
                version,
                branch,
                revision,
                buildno,
                builddate,
                compiler,
            ) = platform._sys_version(input)
            self.assertEqual((name, version, "", "", buildno, builddate, compiler), output)

        # Tests for python_implementation(), python_version(), python_branch(),
        # python_revision(), python_build(), and python_compiler().
        sys_versions = {
            (
                "2.6.1 (r261:67515, Dec  6 2008, 15:26:00) \n[GCC 4.0.1 (Apple Computer, Inc. build 5370)]",
                ("CPython", "tags/r261", "67515"),
                self.save_platform,
            ): (
                "CPython",
                "2.6.1",
                "tags/r261",
                "67515",
                ("r261:67515", "Dec  6 2008 15:26:00"),
                "GCC 4.0.1 (Apple Computer, Inc. build 5370)",
            ),
            ("IronPython 2.0 (2.0.0.0) on .NET 2.0.50727.3053", None, "cli"): (
                "IronPython",
                "2.0.0",
                "",
                "",
                ("", ""),
                ".NET 2.0.50727.3053",
            ),
            (
                "2.6.1 (IronPython 2.6.1 (2.6.10920.0) on .NET 2.0.50727.1433)",
                None,
                "cli",
            ): ("IronPython", "2.6.1", "", "", ("", ""), ".NET 2.0.50727.1433"),
            (
                "2.7.4 (IronPython 2.7.4 (2.7.0.40) on Mono 4.0.30319.1 (32-bit))",
                None,
                "cli",
            ): ("IronPython", "2.7.4", "", "", ("", ""), "Mono 4.0.30319.1 (32-bit)"),
            (
                '2.5 (trunk:6107, Mar 26 2009, 13:02:18) \n[Java HotSpot(TM) Client VM ("Apple Computer, Inc.")]',
                ("Jython", "trunk", "6107"),
                "java1.5.0_16",
            ): (
                "Jython",
                "2.5.0",
                "trunk",
                "6107",
                ("trunk:6107", "Mar 26 2009"),
                "java1.5.0_16",
            ),
            (
                "2.5.2 (63378, Mar 26 2009, 18:03:29)\n[PyPy 1.0.0]",
                ("PyPy", "trunk", "63378"),
                self.save_platform,
            ): ("PyPy", "2.5.2", "trunk", "63378", ("63378", "Mar 26 2009"), ""),
        }
        for (version_tag, scm, sys_platform), info in sys_versions.items():
            sys.version = version_tag
            if scm is None:
                if hasattr(sys, "_git"):
                    del sys._git
            else:
                sys._git = scm
            if sys_platform is not None:
                sys.platform = sys_platform
            self.assertEqual(platform.python_implementation(), info[0])
            self.assertEqual(platform.python_version(), info[1])
            self.assertEqual(platform.python_branch(), info[2])
            self.assertEqual(platform.python_revision(), info[3])
            self.assertEqual(platform.python_build(), info[4])
            self.assertEqual(platform.python_compiler(), info[5])

    def test_system_alias(self):
        res = platform.system_alias(
            platform.system(),
            platform.release(),
            platform.version(),
        )

    def test_uname(self):
        res = platform.uname()
        self.assertTrue(any(res))
        self.assertEqual(res[0], res.system)
        self.assertEqual(res[-6], res.system)
        self.assertEqual(res[1], res.node)
        self.assertEqual(res[-5], res.node)
        self.assertEqual(res[2], res.release)
        self.assertEqual(res[-4], res.release)
        self.assertEqual(res[3], res.version)
        self.assertEqual(res[-3], res.version)
        self.assertEqual(res[4], res.machine)
        self.assertEqual(res[-2], res.machine)
        self.assertEqual(res[5], res.processor)
        self.assertEqual(res[-1], res.processor)
        self.assertEqual(len(res), 6)

    def test_uname_cast_to_tuple(self):
        res = platform.uname()
        expected = (
            res.system,
            res.node,
            res.release,
            res.version,
            res.machine,
            res.processor,
        )
        self.assertEqual(tuple(res), expected)

    def test_uname_replace(self):
        res = platform.uname()
        new = res._replace(
            system="system",
            node="node",
            release="release",
            version="version",
            machine="machine",
        )
        self.assertEqual(new.system, "system")
        self.assertEqual(new.node, "node")
        self.assertEqual(new.release, "release")
        self.assertEqual(new.version, "version")
        self.assertEqual(new.machine, "machine")
        # processor cannot be replaced
        self.assertEqual(new.processor, res.processor)

    def test_uname_copy(self):
        uname = platform.uname()
        self.assertEqual(copy.copy(uname), uname)
        self.assertEqual(copy.deepcopy(uname), uname)

    @unittest.skip("myhrr not supported feature")
    def test_uname_pickle(self):
        orig = platform.uname()
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            with self.subTest(protocol=proto):
                pickled = pickle.dumps(orig, proto)
                restored = pickle.loads(pickled)
                self.assertEqual(restored, orig)

    def test_uname_slices(self):
        res = platform.uname()
        expected = tuple(res)
        self.assertEqual(res[:], expected)
        self.assertEqual(res[:5], expected[:5])

    def test_uname_fields(self):
        self.assertIn("processor", platform.uname()._fields)

    def test_uname_asdict(self):
        res = platform.uname()._asdict()
        self.assertEqual(len(res), 6)
        self.assertIn("processor", res)

    @unittest.skipIf(sys.platform in ["win32", "OpenVMS"], "uname -p not used")
    @support.requires_subprocess()
    def test_uname_processor(self):
        """
        On some systems, the processor must match the output
        of 'uname -p'. See Issue 35967 for rationale.
        """
        try:
            proc_res = subprocess.check_output(["uname", "-p"], text=True).strip()
            expect = platform._unknown_as_blank(proc_res)
        except (OSError, subprocess.CalledProcessError):
            expect = ""
        self.assertEqual(platform.uname().processor, expect)

    @unittest.skipUnless(sys.platform.startswith("win"), "windows only test")
    def test_uname_win32_ARCHITEW6432(self):
        # Issue 7860: make sure we get architecture from the correct variable
        # on 64 bit Windows: if PROCESSOR_ARCHITEW6432 exists we should be
        # using it, per
        # http://blogs.msdn.com/david.wang/archive/2006/03/26/HOWTO-Detect-Process-Bitness.aspx
        try:
            with os_helper.EnvironmentVarGuard() as environ:
                if "PROCESSOR_ARCHITEW6432" in environ:
                    del environ["PROCESSOR_ARCHITEW6432"]
                environ["PROCESSOR_ARCHITECTURE"] = "foo"
                platform._uname_cache = None
                system, node, release, version, machine, processor = platform.uname()
                self.assertEqual(machine, "foo")
                environ["PROCESSOR_ARCHITEW6432"] = "bar"
                platform._uname_cache = None
                system, node, release, version, machine, processor = platform.uname()
                self.assertEqual(machine, "bar")
        finally:
            platform._uname_cache = None

    def test_java_ver(self):
        res = platform.java_ver()
        if sys.platform == "java":
            self.assertTrue(all(res))

    def test_win32_ver(self):
        res = platform.win32_ver()

    def test_mac_ver(self):
        res = platform.mac_ver()

        if platform.uname().system == "Darwin":
            # We are on a macOS system, check that the right version
            # information is returned
            output = subprocess.check_output(["sw_vers"], text=True)
            for line in output.splitlines():
                if line.startswith("ProductVersion:"):
                    real_ver = line.strip().split()[-1]
                    break
            else:
                self.fail(f"failed to parse sw_vers output: {output!r}")

            result_list = res[0].split(".")
            expect_list = real_ver.split(".")
            len_diff = len(result_list) - len(expect_list)
            # On Snow Leopard, sw_vers reports 10.6.0 as 10.6
            if len_diff > 0:
                expect_list.extend(["0"] * len_diff)
            # For compatibility with older binaries, macOS 11.x may report
            # itself as '10.16' rather than '11.x.y'.
            if result_list != ["10", "16"]:
                self.assertEqual(result_list, expect_list)

            # res[1] claims to contain
            # (version, dev_stage, non_release_version)
            # That information is no longer available
            self.assertEqual(res[1], ("", "", ""))

            if sys.byteorder == "little":
                self.assertIn(res[2], ("i386", "x86_64", "arm64"))
            else:
                self.assertEqual(res[2], "PowerPC")

    @unittest.skipUnless(sys.platform == "darwin", "OSX only test")
    def test_mac_ver_with_fork(self):
        # Issue7895: platform.mac_ver() crashes when using fork without exec
        #
        # This test checks that the fix for that issue works.
        #
        pid = os.fork()
        if pid == 0:
            # child
            info = platform.mac_ver()
            os._exit(0)

        else:
            # parent
            support.wait_process(pid, exitcode=0)

    @unittest.skipIf(support.is_emscripten, "Does not apply to Emscripten")
    def test_libc_ver(self):
        # check that libc_ver(executable) doesn't raise an exception
        if os.path.isdir(sys.executable) and os.path.exists(sys.executable + ".exe"):
            # Cygwin horror
            executable = sys.executable + ".exe"
        elif sys.platform == "win32" and not os.path.exists(sys.executable):
            # App symlink appears to not exist, but we want the
            # real executable here anyway
            import _winapi

            executable = _winapi.GetModuleFileName(0)
        else:
            executable = sys.executable
        platform.libc_ver(executable)

        filename = os_helper.TESTFN
        self.addCleanup(os_helper.unlink, filename)

        with mock.patch.object(os, "confstr", create=True, return_value="mock 1.0"):
            # test os.confstr() code path
            self.assertEqual(platform.libc_ver(), ("mock", "1.0"))

            # test the different regular expressions
            for data, expected in (
                (b"__libc_init", ("libc", "")),
                (b"GLIBC_2.9", ("glibc", "2.9")),
                (b"libc.so.1.2.5", ("libc", "1.2.5")),
                (b"libc_pthread.so.1.2.5", ("libc", "1.2.5_pthread")),
                (b"", ("", "")),
            ):
                with open(filename, "wb") as fp:
                    fp.write(b"[xxx%sxxx]" % data)
                    fp.flush()

                # os.confstr() must not be used if executable is set
                self.assertEqual(platform.libc_ver(executable=filename), expected)

        # binary containing multiple versions: get the most recent,
        # make sure that 1.9 is seen as older than 1.23.4
        chunksize = 16384
        with open(filename, "wb") as f:
            # test match at chunk boundary
            f.write(b"x" * (chunksize - 10))
            f.write(b"GLIBC_1.23.4\0GLIBC_1.9\0GLIBC_1.21\0")
        self.assertEqual(platform.libc_ver(filename, chunksize=chunksize), ("glibc", "1.23.4"))

    @support.cpython_only
    def test__comparable_version(self):
        from platform import _comparable_version as V

        self.assertEqual(V("1.2.3"), V("1.2.3"))
        self.assertLess(V("1.2.3"), V("1.2.10"))
        self.assertEqual(V("1.2.3.4"), V("1_2-3+4"))
        self.assertLess(V("1.2spam"), V("1.2dev"))
        self.assertLess(V("1.2dev"), V("1.2alpha"))
        self.assertLess(V("1.2dev"), V("1.2a"))
        self.assertLess(V("1.2alpha"), V("1.2beta"))
        self.assertLess(V("1.2a"), V("1.2b"))
        self.assertLess(V("1.2beta"), V("1.2c"))
        self.assertLess(V("1.2b"), V("1.2c"))
        self.assertLess(V("1.2c"), V("1.2RC"))
        self.assertLess(V("1.2c"), V("1.2rc"))
        self.assertLess(V("1.2RC"), V("1.2.0"))
        self.assertLess(V("1.2rc"), V("1.2.0"))
        self.assertLess(V("1.2.0"), V("1.2pl"))
        self.assertLess(V("1.2.0"), V("1.2p"))

        self.assertLess(V("1.5.1"), V("1.5.2b2"))
        self.assertLess(V("3.10a"), V("161"))
        self.assertEqual(V("8.02"), V("8.02"))
        self.assertLess(V("3.4j"), V("1996.07.12"))
        self.assertLess(V("3.1.1.6"), V("3.2.pl0"))
        self.assertLess(V("2g6"), V("11g"))
        self.assertLess(V("0.9"), V("2.2"))
        self.assertLess(V("1.2"), V("1.2.1"))
        self.assertLess(V("1.1"), V("1.2.2"))
        self.assertLess(V("1.1"), V("1.2"))
        self.assertLess(V("1.2.1"), V("1.2.2"))
        self.assertLess(V("1.2"), V("1.2.2"))
        self.assertLess(V("0.4"), V("0.4.0"))
        self.assertLess(V("1.13++"), V("5.5.kw"))
        self.assertLess(V("0.960923"), V("2.2beta29"))

    def test_macos(self):
        self.addCleanup(self.clear_caches)

        uname = (
            "Darwin",
            "hostname",
            "17.7.0",
            ("Darwin Kernel Version 17.7.0: " "Thu Jun 21 22:53:14 PDT 2018; " "root:xnu-4570.71.2~1/RELEASE_X86_64"),
            "x86_64",
            "i386",
        )
        arch = ("64bit", "")
        with mock.patch.object(platform, "uname", return_value=uname), mock.patch.object(platform, "architecture", return_value=arch):
            for mac_ver, expected_terse, expected in [
                # darwin: mac_ver() returns empty strings
                (("", "", ""), "Darwin-17.7.0", "Darwin-17.7.0-x86_64-i386-64bit"),
                # macOS: mac_ver() returns macOS version
                (
                    ("10.13.6", ("", "", ""), "x86_64"),
                    "macOS-10.13.6",
                    "macOS-10.13.6-x86_64-i386-64bit",
                ),
            ]:
                with mock.patch.object(platform, "mac_ver", return_value=mac_ver):
                    self.clear_caches()
                    self.assertEqual(platform.platform(terse=1), expected_terse)
                    self.assertEqual(platform.platform(), expected)

    def test_freedesktop_os_release(self):
        self.addCleanup(self.clear_caches)
        self.clear_caches()

        if any(os.path.isfile(fn) for fn in platform._os_release_candidates):
            info = platform.freedesktop_os_release()
            self.assertIn("NAME", info)
            self.assertIn("ID", info)

            info["CPYTHON_TEST"] = "test"
            self.assertNotIn("CPYTHON_TEST", platform.freedesktop_os_release())
        else:
            with self.assertRaises(OSError):
                platform.freedesktop_os_release()

    def test_parse_os_release(self):
        info = platform._parse_os_release(FEDORA_OS_RELEASE.splitlines())
        self.assertEqual(info["NAME"], "Fedora")
        self.assertEqual(info["ID"], "fedora")
        self.assertNotIn("ID_LIKE", info)
        self.assertEqual(info["VERSION_CODENAME"], "")

        info = platform._parse_os_release(UBUNTU_OS_RELEASE.splitlines())
        self.assertEqual(info["NAME"], "Ubuntu")
        self.assertEqual(info["ID"], "ubuntu")
        self.assertEqual(info["ID_LIKE"], "debian")
        self.assertEqual(info["VERSION_CODENAME"], "focal")

        info = platform._parse_os_release(TEST_OS_RELEASE.splitlines())
        expected = {
            "ID": "linux",
            "NAME": "Linux",
            "PRETTY_NAME": "Linux",
            "ID_LIKE": "egg spam viking",
            "EMPTY": "",
            "DOUBLE_QUOTE": "double",
            "EMPTY_DOUBLE": "",
            "SINGLE_QUOTE": "single",
            "EMPTY_SINGLE": "",
            "QUOTES": "double's",
            "SPECIALS": "$`\\'\"",
        }
        self.assertEqual(info, expected)
        self.assertEqual(len(info["SPECIALS"]), 5)


if __name__ == "__main__":
    unittest.main()



import warnings


from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.core.services.system import AbcRuntime
from collections import namedtuple

from ._mosenv import OsEnv


__mlib__ = "AbcGrp"

struct_group = namedtuple("struct_group", "gr_name gr_passwd gr_gid gr_mem")


class AbcGrp(AbcRuntime):
    __frameworkpath__ = "mpython.mgrp"

    __all__ = ["getgrgid", "getgrnam", "getgrall"]

    struct_group = struct_group

    def __init__(self, system):
        self._os = OsEnv(self)

    def getgrgid(self, gid):
        if not isinstance(gid, int):
            warnings.warn("should be integer", DeprecationWarning)
            gid = int(gid)

        grall = self.getgrall()
        for gr in grall:
            if gr.gr_gid == gid:
                return gr
        raise KeyError("getgrgid(): gid not found: %s" % gid)

    def getgrnam(self, name):
        if not isinstance(name, str):
            raise TypeError("must be string, not %s" % name.__class__.__name__)

        grall = self.getgrall()
        for gr in grall:
            if gr.gr_name == name:
                return gr
        raise KeyError("getgrnam(): name not found: %s" % name)

    def getgrall(self):
        out, err, rval = self.myrrh_os.cmd(b"%(cat)s /etc/group")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        grall = []
        for line in out.split(self._os.linesep):
            try:
                line = line.split(":")
                line[2] = int(line[2])
                line[3] = [u for u in line[3].split(",") if len(u)]
                grall.append(self.struct_group(*line))
            except (TypeError, ValueError, IndexError):
                pass

        return grall



import unittest

from myrrh.core.services.groups import MyrrhGroup, myrrh_group


class TestGroup(unittest.TestCase):
    def test_basic(self):
        g = MyrrhGroup("a", "b")
        self.assertIsInstance(g, MyrrhGroup)

    def test_basic_attr(self):
        g = MyrrhGroup(",", ";")

        gs = g.join("abc")
        gas = g._as_.join("abc")

        self.assertEqual(gs._t_, gas._t_)

    def test_basic_func(self):
        def replace1(s):
            return s.replace("1", "[one]")

        g = MyrrhGroup("2354891251478", "2354891251478", keys=("a", "b"))

        gs = myrrh_group(replace1)(g)
        gas = myrrh_group(replace1)._as_(g)

        self.assertEqual(gs._t_, gas._t_)

    def test_basic_iter(self):
        g = MyrrhGroup("012345678", "abcdefghi", keys=("num", "string"))
        i0, i1 = iter("012345678"), iter("abcdefghi")

        for v in g:
            self.assertEqual(v._t_, (next(i0), next(i1)))

        g = MyrrhGroup(0, 1, keys=("a", "b"))
        self.assertRaises(TypeError, iter, g)

    def test_basic_subscript(self):
        g = MyrrhGroup(list("012345678"), list("abcdefghi"), keys=("num", "string"))
        self.assertEqual(g[0]._t_, ("0", "a"))

        g[0] = "x"
        self.assertEqual(g[0]._t_, ("x", "x"))

        g = MyrrhGroup(1, 2, keys=("num", "string"))
        self.assertRaises(TypeError, lambda: g[0])


if __name__ == "__main__":
    unittest.main()



import re
import posixpath

from myrrh.core.services.system import AbcMyrrhOs, ExecutionFailureCauseRVal
from myrrh.utils import mshlex
from myrrh.utils import merrno as errno


class MyrrhOs(AbcMyrrhOs):
    _curdirb_ = b"."
    _pardirb_ = b".."
    _extsepb_ = b"."
    _sepb_ = b"/"
    _pathsepb_ = b":"
    _defpathb_ = b":/system/bin"
    _altsepb_ = None
    _devnullb_ = b"/dev/null"
    _linesepb_ = b"\n"

    isabs = staticmethod(posixpath.isabs)  # type: ignore[assignment]
    normpath = staticmethod(posixpath.normpath)  # type: ignore[assignment]
    joinpath = staticmethod(posixpath.join)  # type: ignore[assignment]
    basename = staticmethod(posixpath.basename)  # type: ignore[assignment]
    dirname = staticmethod(posixpath.dirname)  # type: ignore[assignment]

    def _getbinb_(self):
        return {
            b"find": b"/system/bin/find",
            b"ln": b"/system/bin/ln",
            b"chown": b"/system/bin/chown",
            b"chgrp": b"/system/bin/chgrp",
            b"chmod": b"/system/bin/chmod",
            b"mv": b"/system/bin/mv",
            b"realpath": b"command realpath",
            b"touch": b"/system/bin/touch",
            b"cat": b"/system/bin/cat",
            b"pwd": b"command pwd",
            b"sh": b"/system/bin/sh",
            b"set": b"set",
            b"which": b"command -v",
            b"dirname": b"/system/bin/dirname",
            b"id": b"/system/bin/id",
            b"stat": b"/system/bin/stat",
            b"echo": b"print",
            b"truncate": b"/system/bin/truncate",
            b"sleep": b"/system/bin/sleep",
            b"ps": b"/system/bin/ps",
            b"xargs": b"/system/bin/xargs",
            b"getprop": b"/system/bin/getprop",
            b"cp": b"/system/bin/cp",
            b"tar": b"/system/bin/tar",
        }

    def _getdefaultshellb_(self):
        return self.getenvb().get(b"SHELL", self._getbinb_()[b"sh"])

    def _getdefaultshellargsb_(self):
        return (b"-c",)

    def formatshellargs(self, args, *, defaultargs=None):
        if isinstance(args, str):
            args = args.encode()
        return (
            *defaultargs,
            mshlex.list2cmdlineb(args).strip(b'"') if isinstance(args, (list, tuple)) else args,
        )

    def _getdefaultencoding_(self):
        return "utf-8"

    def _fsencoding_(self):
        return "utf-8"

    def _fsencodeerrors_(self):
        return "surrogateescape"

    def sh_escape_bytes(self, string):
        return mshlex.quote(string)

    def _localecode_(self):
        return ""

    def _getcwdb_(self):
        out, err, rval = self.cmdb(b"%(pwd)s")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out

    def _getreadonlyenvb_(self):
        return (b"PPID", b"KSH_VERSION", b"PIPESTATUS")

    def _envb_(self):
        out, err, rval = self.cmdb(b"%(set)s")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        env = {k: v for k, v in re.findall(b"(?P<k>[^\\n=]*)=(?P<v>'[^']*'\\n|.*\\n)", out)}
        env = {k: v.rstrip(b"\n") for k, v in env.items()}
        return env

    def _gettmpdirb_(self):
        out, err, rval = self.cmdb(b'%(echo)s "$TMPDIR:$TEMP:$TMP:`[ -d /tmp ] && %(echo)s /tmp`:`[ -d /var/tmp ] && %(echo)s /var/tmp`:`[ -d /usr/tmp ] && %(echo)s /usr/tmp`"')
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        dirlist = [dir for dir in filter(None, out.split(b":"))]
        return dirlist[0] if dirlist else self._getcwdb_()

    def default_errno_from_msg(self, err):
        err = self.shencode(err)
        return errno.errno_from_msgb(err)

    environkeyformat = None  # type: ignore[assignment]



from myrrh.core.services.system import AbcRuntime


class Ansimap(AbcRuntime):
    CSI = {"\r": "\r\n"}

    def __init__(self, control, iterator, abortkey=None):
        self.map = self.__posixtowin
        self.__iter = iterator
        self.__abortkey = abortkey
        self.__keys = []

    def __iter__(self):
        return self

    def __next__(self):
        return self.map()

    def __call__(self):
        return self.map()

    def __posixtowin(self):
        if len(self.__keys) == 0:
            k = next(self.__iter)
            self.__keys.extend(self.CSI.get(k, k))

        return self.__keys.pop(0)



import unittest

from myrrh.provider import IProvider, IShellService, IFileSystemService
from myrrh.provider.registry import ProviderRegistry

from myrrh.core.services.entity import CoreProvider


class CoreProviderTest(unittest.TestCase):
    def test_basic_create(self):
        ProviderRegistry()._new_provider("new", ProviderRegistry().local)
        p_local = ProviderRegistry().get("local")

        p = CoreProvider((p_local(),))

        self.assertEqual(len(p_local().services()), len(p.services()))
        self.assertEqual(set(p_local().catalog()), set(p.catalog()))

    def test_basic_set_default(self):
        class FakeProvider(IProvider):
            class FakeShellServ(IShellService):
                protocol = "fake"

            class FakeFilesystemServ(IFileSystemService):
                protocol = "fake"

            def services(self):
                return (self.FakeShellServ, self.FakeFilesystemServ)

            def catalog(self):
                return tuple()

            def deliver(self, name):
                ...

        ProviderRegistry()._new_provider("fake", FakeProvider)
        p_local = ProviderRegistry().get("local")()
        p_fake = ProviderRegistry().get("fake")()

        p = CoreProvider((p_local, p_fake))
        self.assertEqual(len(p.services()), len(p_fake.services() + p_local.services()))

        p = CoreProvider((p_local, p_fake), patterns=("**/local",))
        self.assertEqual(len(p.services()), len(p_local.services()))
        p = CoreProvider((p_local, p_fake), patterns=("*system*/fake",))
        self.assertEqual(len(p.services()), len(p_fake.services()))


if __name__ == "__main__":
    unittest.main()



import threading

from myrrh.utils import mshlex

from myrrh.utils import mstring
from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.core.services.system import AbcRuntime
from myrrh.framework.mpython import mimportlib


from myrrh.resources.textfsm import getparser

__mlib__ = "WinReg"


def _convert_reg_sz(v):
    v = str(v)
    return v[: v.index("\0")] if "\0" in v else v


def _convert_reg_multi_sz(v):
    return "\\0".join(filter(None, v))


def _revert_reg_multi_sz(v):
    return v.split("\\0") if v else []


class WinReg(AbcRuntime):
    """
    Notice : - RE_MULTI_SZ does not support empty string
             - Big data unsupported
    """

    __frameworkpath__ = "mpython.winreg"

    tempfile = mimportlib.module_property("tempfile")
    os = mimportlib.module_property("os")

    QUERY_PARSER = staticmethod(getparser("nt_reg"))

    def __init__(self, system=None):
        self._hkeys = dict()

        self.HKEY_CLASSES_ROOT = self.PyHKEY(0, self, None, "HKEY_CLASSES_ROOT")
        self.HKEY_CURRENT_USER = self.PyHKEY(0, self, None, "HKEY_CURRENT_USER")
        self.HKEY_LOCAL_MACHINE = self.PyHKEY(0, self, None, "HKEY_LOCAL_MACHINE")
        self.HKEY_USERS = self.PyHKEY(0, self, None, "HKEY_USERS")
        self.HKEY_PERFORMANCE_DATA = self.PyHKEY(0, self, None, "HKEY_PERFORMANCE_DATA")
        self.HKEY_CURRENT_CONFIG = self.PyHKEY(0, self, None, "HKEY_CURRENT_CONFIG")
        self.HKEY_DYN_DATA = self.PyHKEY(0, self, None, "HKEY_DYN_DATA")

        self._default_key = None

    @property
    def default_key(self):
        if self._default_key is None:
            self._default_key, _, _ = self.HKEY_CURRENT_USER._query_default()

    class PyHKEY:
        def __init__(self, handle, winreg, hkey, key="", *, computer=None):
            self.handle = handle
            self.winreg = winreg
            self.myrrh_os = winreg.myrrh_os

            self._closed = False
            self._detached = False

            self._lock = threading.Lock()
            self.__subkeys = None
            self.__values = None

            self._computer = computer or (hkey and hkey._computer) or "."

            if hkey:
                self.key = self._join(hkey.key, key) if key else hkey.key
            else:
                self.key = key

        @property
        def _values(self):
            if self.__values is None:
                self._query()

            return self.__values

        @property
        def _subkeys(self):
            if self.__subkeys is None:
                self._query()

            return self.__subkeys

        def __del__(self):
            if not self._detached:
                self.Detach()

        def __int__(self):
            return self.handle

        def __bool__(self):
            return not self._closed

        def __enter__(self):
            return self

        def __exit__(self, *exec_info):
            if not self._detached:
                self.Close()

        def _join(self, *a):
            return "\\".join(a)

        @property
        def path(self):
            return self._join("\\", self._computer, self.key)

        def Detach(self):
            if self.handle and not self._detached:
                self.winreg._pop(self.handle)

            self._detached = True
            return self.handle

        def Close(self):
            self.Detach()
            self._closed = True
            self.handle = 0

        def _new(self, handle, sub_key=None):
            if not (sub_key):
                return self
            return WinReg.PyHKEY(handle, self.winreg, self, sub_key, computer=self._computer)

        def cmd(self, *a, **kwa):
            return self.myrrh_os.cmd(*a, **kwa)

        def sh_escape_bytes(self, bytes_or_str):
            value = self.myrrh_os.fsencode(bytes_or_str)
            return mshlex.winshell_escape_for_stringyfication_in_cmdb(value)

        def _find_value_index(self, name):
            for a in self._values:
                n, _, _ = a
                if n == name:
                    return self._values.index(a)

        def _add_value(self, value):
            idx = self._find_value_index(value[0])
            if idx is None:
                self._values.append(value)
            else:
                self._values[idx] = value

        def _query_default(self):
            key_path = self.myrrh_os.p(self.path)

            o, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s QUERY "%(path)s" /ve',
                path=self.myrrh_os.sh_escape_bytes(key_path),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, key_path).check()

            with self.winreg.QUERY_PARSER as textfsm:
                fsm = textfsm.ParseText(o)

            return fsm[0][1:] if len(fsm[0]) else tuple()

        def _query(self, o=None):
            key_path = self.myrrh_os.p(self.path)

            if o is None:
                o, e, r = self.myrrh_os.cmd(
                    b'%(reg_utf8)s QUERY "%(path)s"',
                    path=self.myrrh_os.sh_escape_bytes(key_path),
                )
                try:
                    ExecutionFailureCauseRVal(self, e, r, 0, key_path).check()
                except ExecutionFailureCauseRVal:
                    raise FileNotFoundError(key_path)

            with self._lock:
                with self.winreg.QUERY_PARSER as textfsm:
                    fsm = textfsm.ParseText(o)

                self.__subkeys = list()
                self.__values = list()
                for line in fsm:
                    k, n, v, t = line

                    if k != self.key:
                        k = self.winreg.os.path.basename(k)

                        try:
                            self._subkeys.index(k)
                        except ValueError:
                            self._subkeys.append(k)

                    else:
                        if not t:
                            continue

                        if n == self.winreg.default_key:
                            n = ""

                        stype, value_conv = self.winreg._TYPE_CONVERT[t]
                        value = value_conv(v)

                        self._add_value((n, value, stype))

        def _create(self):
            o, e, r = self.myrrh_os.cmd(
                b'( %(reg_utf8)s QUERY "%(path)s" 2>NUL ) || ( ( %(reg_utf8)s ADD "%(path)s" 1>NUL ) && ( %(reg_utf8)s DELETE "%(path)s" /ve /f 1>NUL ) && %(reg_utf8)s QUERY "%(path)s")',
                path=self.myrrh_os.sh_escape_bytes(self.key_path),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()
            self._query(o)

        def _delete(self):
            key_path = self.myrrh_os.p(self.path)

            _, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s DELETE "%(path)s" /f',
                path=self.myrrh_os.sh_escape_bytes(key_path),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()

        def _delete_value(self, value):
            key_path = self.myrrh_os.p(self.path)
            value = self.myrrh_os.shdecode(value)

            o, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s DELETE  "%(path)s" /v "%(value)s" /f && %(reg_utf8)s QUERY "%(path)s"',
                path=self.myrrh_os.sh_escape_bytes(key_path),
                value=self.myrrh_os.sh_escape_bytes(value),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()
            self._query(o)

        def _loadkey(self, filename):
            key_path = self.myrrh_os.p(self.path)
            filename = self.myrrh_os.p(filename)

            o, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s LOAD "%(path)s" "%(filename)s" && %(reg_utf8)s QUERY "%(path)s"',
                path=self.myrrh_os.sh_escape_bytes(key_path),
                value=self.myrrh_os.sh_escape_bytes(filename),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()
            self._query(o)

        def _save(self, filename):
            key_path = self.myrrh_os.p(self.path)
            filename = self.myrrh_os.p(filename)

            _, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s SAVE "%(path)s" "%(filename)s"',
                path=self.myrrh_os.sh_escape_bytes(key_path),
                value=self.myrrh_os.sh_escape_bytes(filename),
            )
            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()

        def _set_value(self, type, value, value_name=""):
            key_path = self.myrrh_os.p(self.path)

            stype, value_conv = self.winreg._TYPE_CONVERT[type]
            svalue = value_conv(value) if value is not None else ""

            stype = self.myrrh_os.shdecode(value)
            svalue = self.myrrh_os.shdecode(value)

            o, e, r = self.myrrh_os.cmd(
                b'%(reg_utf8)s ADD "%(path)s"  %(value_name)s /t %(type)s /d "%(value)s" /f && %(reg_utf8)s QUERY "%(path)s" ',
                path=self.myrrh_os.sh_escape_bytes(key_path),
                value=self.myrrh_os.sh_escape_bytes(svalue),
                type=self.myrrh_os.sh_escape_bytes(stype),
                value_name=(b'/v "%s"' % self.myrrh_os.sh_escape_bytes(value_name)) if value_name else b"",
            )

            ExecutionFailureCauseRVal(self, e, r, 0, self.path).check()

            self._query(o)

    HKEYType = PyHKEY

    HKEY_CLASSES_ROOT: PyHKEY | None = None
    HKEY_CURRENT_USER: PyHKEY | None = None
    HKEY_LOCAL_MACHINE: PyHKEY | None = None
    HKEY_USERS: PyHKEY | None = None
    HKEY_PERFORMANCE_DATA: PyHKEY | None = None
    HKEY_CURRENT_CONFIG: PyHKEY | None = None
    HKEY_DYN_DATA: PyHKEY | None = None

    REG_NONE = 0
    REG_SZ = 1
    REG_EXPAND_SZ = 2
    REG_BINARY = 3
    REG_DWORD = 4
    REG_DWORD_LITTLE_ENDIAN = 4
    REG_DWORD_BIG_ENDIAN = 5
    REG_LINK = 6
    REG_MULTI_SZ = 7
    REG_RESOURCE_LIST = 8
    REG_FULL_RESOURCE_DESCRIPTOR = 9
    REG_RESOURCE_REQUIREMENTS_LIST = 10
    REG_QWORD_LITTLE_ENDIAN = 11
    REG_QWORD = 11

    _TYPE_CONVERT = {
        "REG_NONE": (REG_NONE, str),
        "REG_SZ": (REG_SZ, str),
        "REG_EXPAND_SZ": (REG_EXPAND_SZ, str),
        "REG_BINARY": (REG_BINARY, lambda v: bytes.fromhex(v) if v != "" else None),
        "REG_DWORD": (REG_DWORD, mstring.str2int),
        "REG_MULTI_SZ": (REG_MULTI_SZ, _revert_reg_multi_sz),
        "REG_QWORD": (REG_QWORD, mstring.str2int),
        REG_NONE: ("REG_NONE", str),
        REG_SZ: ("REG_SZ", _convert_reg_sz),
        REG_EXPAND_SZ: ("REG_EXPAND_SZ", str),
        REG_BINARY: ("REG_BINARY", lambda v: v.hex() if v is not None else ""),
        REG_DWORD: ("REG_DWORD", str),
        REG_DWORD_BIG_ENDIAN: ("REG_DWORD", str),
        REG_LINK: ("REG_SZ", str),
        REG_MULTI_SZ: ("REG_MULTI_SZ", _convert_reg_multi_sz),
        REG_RESOURCE_LIST: ("REG_MULTI_SZ", str),
        REG_FULL_RESOURCE_DESCRIPTOR: ("REG_SZ", str),
        REG_RESOURCE_REQUIREMENTS_LIST: ("REG_MULTI_SZ", str),
        REG_QWORD: ("REG_QWORD", str),
    }

    KEY_ALL_ACCESS = 0xF003F
    KEY_QUERY_VALUE = 0x1
    KEY_SET_VALUE = 0x2
    KEY_CREATE_SUB_KEY = 0x4
    KEY_ENUMERATE_SUB_KEYS = 0x8
    KEY_NOTIFY = 0x10
    KEY_CREATE = 0x20
    KEY_WRITE = 0x20006
    KEY_READ = 0x20019
    KEY_EXECUTE = 0x20019
    KEY_CREATE_LINK = 0x20

    KEY_WOW64_64KEY = 0x0100
    KEY_WOW64_32KEY = 0x0200

    DEFAULT_NEW_KEY = "new key"

    REG_CREATED_NEW_KEY = 1
    REG_REFRESH_HIVE = 2

    REG_WHOLE_HIVE_VOLATILE = 1

    REG_LEGAL_CHANGE_FILTER = 0x1000000F
    REG_LEGAL_OPTION = 0x1F

    REG_NOTIFY_CHANGE_NAME = 1
    REG_NOTIFY_CHANGE_ATTRIBUTES = 2
    REG_NOTIFY_CHANGE_LAST_SET = 4
    REG_NOTIFY_CHANGE_SECURITY = 8

    REG_OPENED_EXISTING_KEY = 2

    REG_OPTION_RESERVED = 0
    REG_OPTION_NON_VOLATILE = 0
    REG_OPTION_VOLATILE = 1
    REG_OPTION_CREATE_LINK = 2
    REG_OPTION_BACKUP_RESTORE = 4
    REG_OPTION_OPEN_LINK = 8

    REG_NO_LAZY_FLUSH = 4

    _START_HANDLE = 777

    def _to_PyHKEY(self, handle):
        if isinstance(handle, int):
            try:
                return self._hkeys[handle]
            except KeyError:
                pass

            raise OSError(6, "Invalid descriptor")

        if isinstance(handle, WinReg.PyHKEY):
            return handle

        raise TypeError("The object is not a PyHKEY object")

    def _to_RefHKEY(self, handle):
        if isinstance(handle, int):
            try:
                return self._hkeys[handle].handle
            except KeyError:
                pass

            raise OSError(6, "Invalid descriptor")

        if isinstance(handle, WinReg.PyHKEY):
            return handle.handle

        raise TypeError("The object is not a PyHKEY object")

    def _new_HANDLE(self, handle=None):
        if not handle:
            handle = self._START_HANDLE
            self._START_HANDLE += 1
        return handle

    def _close(self, handle):
        hkey = self._to_PyHKEY(handle)
        hkey.Close()

    def _pop(self, hkey):
        handle = self._to_RefHKEY(hkey)
        return self._hkeys.pop(handle)

    def _add(self, hkey):
        self._hkeys[hkey.handle] = hkey
        return hkey

    def CloseKey(self, hkey):
        hkey = self._to_PyHKEY(hkey)
        hkey.Close()

    def ConnectRegistry(self, computer_name, key):
        hkey = self._to_PyHKEY(key)
        hkey = self.PyHKEY(self._new_HANDLE(), self, hkey, computer=computer_name)

        hkey._query()

        return self._add(hkey)

    def CreateKey(self, key, sub_key):
        hkey = self._to_PyHKEY(key)
        hkey = hkey._new(self._new_HANDLE(), sub_key)

        hkey._create()

        return self._add(hkey)

    def CreateKeyEx(self, key, sub_key, reserved=0, access=KEY_WRITE):
        # Access is not used in Myrrh
        return self.CreateKey(key, sub_key)

    def DeleteKey(self, key, sub_key):
        hkey = self._to_PyHKEY(key)
        hkey = hkey._new(0, sub_key)
        hkey._delete()

    def DeleteKeyEx(self, key, sub_key, access=KEY_WOW64_64KEY, reserved=0):
        # Access is not used in Myrrh
        return self.DeleteKey(key, sub_key)

    def DeleteValue(self, key, value):
        hkey = self._to_PyHKEY(key)
        hkey._delete_value(value)

    def EnumKey(self, key, index):
        hkey = self._to_PyHKEY(key)
        try:
            return hkey._subkeys[index]
        except IndexError:
            pass

        raise OSError(259, "No data found")

    def EnumValue(self, key, index):
        hkey = self._to_PyHKEY(key)

        try:
            return hkey._values[index]
        except IndexError:
            pass

        raise OSError(259, "No data found")

    def ExpandEnvironmentStrings(self, str):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s %(str)s", str=self.myrrh_os.shencode(str))
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def FlushKey(self, key):
        hkey = self._to_PyHKEY(key)
        hkey._query()

    def LoadKey(self, key, sub_key, file_name):
        hkey = self._to_PyHKEY(key)
        hkey = hkey._new(0, sub_key)
        hkey._loadkey(file_name)

    def OpenKey(self, key, sub_key, reserved=0, access=KEY_READ):
        hkey = self._to_PyHKEY(key)
        hkey = hkey._new(0, sub_key)
        hkey._query()
        hkey.handle = self._new_HANDLE()
        self._add(hkey)
        return hkey

    def OpenKeyEx(self, key, sub_key, reserved=0, access=KEY_READ):
        return self.OpenKey(key, sub_key, reserved, access)

    def QueryInfoKey(self, key):
        hkey = self._to_PyHKEY(key)
        hkey._query()
        return (len(hkey._subkeys), len(hkey._values), 0)

    def QueryValue(self, key, sub_key):
        hkey = self._to_PyHKEY(key)
        nkey = hkey._new(0, sub_key)
        if nkey is not hkey:
            nkey._query()
            hkey = nkey
        idx = hkey._find_value_index("")
        if idx is None:
            raise FileNotFoundError(key)

        return hkey._values[idx][1]

    def QueryValueEx(self, key, value_name):
        hkey = self._to_PyHKEY(key)
        idx = hkey._find_value_index(value_name)
        if idx is None:
            raise FileNotFoundError(key)
        return hkey._values[idx][1:]

    def SaveKey(self, key, file_name):
        hkey = self._to_PyHKEY(key)
        hkey._save(file_name)

    def SetValue(self, key, sub_key, type, value):
        hkey = self._to_PyHKEY(key)

        if not sub_key:
            sub_key = self.DEFAULT_NEW_KEY
            inc = 1
            while sub_key in hkey._subkeys:
                sub_key = self.DEFAULT_NEW_KEY + " #%d" % inc
                inc += 1

        nkey = hkey._new(0, sub_key)
        nkey._set_value(type, value)

    def SetValueEx(self, key, value_name, reserved, type, value):
        hkey = self._to_PyHKEY(key)
        hkey._set_value(type, value, value_name)

    def DisableReflectionKey(self, key):
        raise NotImplementedError

    def EnableReflectionKey(self, key):
        raise NotImplementedError

    def QueryReflectionKey(self, key):
        raise NotImplementedError



from abc import abstractmethod
import errno
import stat

from myrrh.framework.mpython import mbuiltins

from myrrh.core.services.system import (
    ExecutionFailureCauseRVal,
    MOsError,
    AbcRuntimeDelegate,
)
from myrrh.core.interfaces import ABC

from myrrh.framework.mpython._mosfs import AbcOsFs, stat_result

__mlib__ = "OsFs"


def _stat_out_to_struct(out):
    stat_list = [int(v, 0) for v in out.strip().split(b",")]
    stat_list[-3] = float(stat_list[-3])
    stat_list[-2] = float(stat_list[-2])
    stat_list[-1] = float(stat_list[-1])
    return stat_result(
        stat_list,
        {
            "st_atime_ns": stat_list[-3] * 1000000000,
            "st_mtime_ns": stat_list[-2] * 1000000000,
            "st_ctime_ns": stat_list[-1] * 1000000000,
        },
    )


def _statvfs_out_to_struct(out):
    stat_list = [int(v, 0) for v in out.strip().split(b",")]
    return OsFs.statvfs_result(*stat_list)


class _interface(ABC):
    import posixpath as local_path

    @abstractmethod
    def normcase(self, s):
        ...

    @abstractmethod
    def splitdrive(self, p):
        ...

    @abstractmethod
    def split(self, p):
        ...

    @abstractmethod
    def samestat(self, s1, s2):
        ...

    @abstractmethod
    def expandvars(self, path):
        ...


class OsFs(_interface, AbcOsFs, AbcRuntimeDelegate):
    _umask = 0o022

    __delegated__ = {_interface: _interface.local_path}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_path, self, {"os": self})

        self.__delegate__(_interface, mod)

    def link(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None, follow_symlinks=True):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _, err, rval = self.myrrh_os.cmd(
            b"%(ln)s %(follow)s %(src)s %(dst)s",
            follow=b"" if follow_symlinks else b"-n",
            src=self.myrrh_os.sh_escape_bytes(_src),
            dst=self.myrrh_os.sh_escape_bytes(_dst),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            src,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def chown(self, path, uid, gid, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        _path = self.myrrh_os.p(path, dir_fd=dir_fd)

        if not isinstance(uid, int):
            raise TypeError("invalid user type")
        if not isinstance(gid, int):
            raise TypeError("invalid group type")

        if uid == -1 and gid == -1:
            return

        exe = b"%(chown)s" if (uid != -1) else b"%(chgrp)s"
        symlink = b"" if follow_symlinks else b"-h"
        user = b"'%i'" % uid if uid != -1 else b""
        group = b"'%i'" % gid if gid != -1 else b""

        _, err, rval = self.myrrh_os.cmdb(
            b"%s %s %s%s%s %s"
            % (
                exe,
                symlink,
                user,
                b":" if uid != -1 and gid != -1 else b"",
                group,
                self.myrrh_os.sh_escape_bytes(_path),
            )
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def chmod(self, path, mode, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        # follow_symlinks is ignore
        _path = self.myrrh_os.p(path, dir_fd=dir_fd)
        mode = stat.S_IMODE(mode)
        _, err, rval = self.myrrh_os.cmdb(
            b"%(chmod)s %(mode)o %(path)s",
            mode=mode,
            path=self.myrrh_os.sh_escape_bytes(_path),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def rename(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _, err, rval = self.myrrh_os.cmdb(
            b"%(mv)s %(src)s %(dst)s",
            src=self.myrrh_os.sh_escape_bytes(_src),
            dst=self.myrrh_os.sh_escape_bytes(_dst),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            src,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def replace(self, src, dst, *args, src_dir_fd=None, dst_dir_fd=None):
        self.rename(src, dst, *args, src_dir_fd=src_dir_fd, dst_dir_fd=dst_dir_fd)

    def listdir(self, path="."):
        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.f(path)

        if not self.isdir(path):
            if not self.exists(path):
                MOsError(self, errno.ENOENT, "No such file or directory", args=(path,)).raised()
            MOsError(self, errno.ENOTDIR, "Not a directory", args=(path,)).raised()
        out, err, rval = self.myrrh_os.cmdb(
            b"%(find)s %(path)s/ -maxdepth 1 -mindepth 1 ! -path %(path)s -perm /777 -print",
            path=self.myrrh_os.sh_escape_bytes(_path),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()
        return [_cast_(self.myrrh_os.basename(f.strip())) for f in out.split(self.myrrh_os.linesepb) if len(f) != 0]

    def _scandir_list(self, path="."):
        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.p(path)
        command = b"%(find)s %(path)s -maxdepth 1 ! -path %(path)s -exec %(stat)s -L -c %%n,0x%%f,%%i,%%d,%%h,%%u,%%g,%%s,%%X,%%Y,%%Z {} \\; || %(stat)s -L -c %%n,0x%%f,%%i,%%d,%%h,%%u,%%g,%%s,%%X,%%Y,%%Z %(path)s"
        out, err, rval = self.myrrh_os.cmdb(command, path=self.myrrh_os.sh_escape_bytes(_path))
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

        result = []
        out = [o.split(b",", 1) for o in filter(None, out.split(b"\n"))]
        for name, fstat in out:
            fname = self.myrrh_os.basename(name)
            fpath = self.myrrh_os.joinpath(_path, fname)
            fstat = _stat_out_to_struct(fstat)
            fmode = stat.filemode(fstat.st_mode)
            if "r" in fmode or "w" in fmode or "x" in fmode:
                result.append(self.DirEntry(_cast_(fname), _cast_(fpath), fstat, self.lstat))

        return result

    def _stat(self, path, *, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        command = b"%(stat)s %(follow)s -c 0x%%f,%%i,%%d,%%h,%%u,%%g,%%s,%%X,%%Y,%%Z %(path)s"

        out, err, rval = self.myrrh_os.cmdb(
            command,
            follow=b"-L" if follow_symlinks else b"",
            path=self.myrrh_os.sh_escape_bytes(_path),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()
        return _stat_out_to_struct(out)

    def _lstat(self, path, *, dir_fd=None):
        return self._stat(path, dir_fd=dir_fd, follow_symlinks=False)

    def _statvfs(self, path):
        _path = self.myrrh_os.f(path)

        command = b"%(stat)s -f -c %%s,%%S,%%b,%%f,%%a,%%c,%%d,%%d,-1,%%l %(path)s"
        out, err, rval = self.myrrh_os.cmdb(command, path=self.myrrh_os.sh_escape_bytes(_path))
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

        return _statvfs_out_to_struct(out)

    def realpath(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        path = self.myrrh_os.p(path)

        out, err, rval = self.myrrh_os.cmdb(b"%(realpath)s %(path)s", path=self.myrrh_os.sh_escape_bytes(path))
        ExecutionFailureCauseRVal(self, err, rval, 0, error_translate=self.myrrh_os.default_errno_from_msg).check()

        return _cast_(out.strip())

    def ismount(self, path):
        path = self.myrrh_os.p(path)
        # copy from posixpath
        if self.islink(path):
            # A symlink can never be a mount point
            return False
        try:
            s1 = self._lstat(path)
            s2 = self._lstat(self.myrrh_os.joinpath(path, self.myrrh_os.pardirb))
        except ExecutionFailureCauseRVal:
            return False  # It doesn't exist -- so not a mount point :-)
        dev1 = s1.st_dev
        dev2 = s2.st_dev
        if dev1 != dev2:
            return True  # path/.. on a different device as path
        ino1 = s1.st_ino
        ino2 = s2.st_ino
        if ino1 == ino2:
            return True  # path/.. is the same i-node as path
        return False

    def _utime(self, path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        if times is None and ns is None:
            _, err, rval = self.myrrh_os.cmdb(
                b"%(touch)s -c -a -m %(follow)s %(path)s",
                follow=b"-h" if follow_symlinks else b"",
                path=self.myrrh_os.sh_escape_bytes(_path),
            )

        else:
            atime, mtime = times or (t * 1e-9 for t in ns)
            _, err, rval = self.myrrh_os.cmdb(
                b"%(touch)s -c %(follow)s -a -d @%(atime)i %(path)s && %(touch)s -c %(follow)s -m -d @%(mtime)i %(path)s",
                follow=b"-h" if follow_symlinks else b"",
                atime=atime,
                mtime=mtime,
                path=self.myrrh_os.sh_escape_bytes(_path),
            )

        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def umask(self, mask):
        last_mask = self._umask
        self._umask = mask
        return last_mask

    def truncate(self, path, length):
        _path = self.myrrh_os.p(path)
        _, err, rval = self.myrrh_os.cmdb(
            b"%(stat)s %(path)s > /dev/null && %(truncate)s -c -s %(length)i %(path)s",
            length=length,
            path=self.myrrh_os.sh_escape_bytes(_path),
        )
        ExecutionFailureCauseRVal(
            self,
            err,
            rval,
            0,
            path,
            error_translate=self.myrrh_os.default_errno_from_msg,
        ).check()

    def samefile(self, path1, path2):
        stat1 = self.stat(path1)
        stat2 = self.stat(path2)

        return self.samestat(stat1, stat2)

    @property
    def environ(self):
        return self.myrrh_os.getenv()
